@@ -10,6 +10,7 @@ import torch.nn.functional as F
 from einops import rearrange
 
 from .rope import RotaryEmbedding, apply_rotary_emb
+from .linear import AnchoredLinear
 
 
 class MHAttention(nn.Module):
@@ -221,7 +222,7 @@ class MHAttention(nn.Module):
         return self.out_proj(output_without_heads)
 
 
-class FeedforwardLayer(nn.Module):
+class FeedforwardBlock(nn.Module):
     """
     ...
     """
@@ -235,6 +236,7 @@ class FeedforwardLayer(nn.Module):
         activation_kwargs=None,
         dropout=0.0,
         linear_module=nn.Linear,
+        raw_input=False,
     ):
         super().__init__()
 
@@ -245,19 +247,25 @@ class FeedforwardLayer(nn.Module):
 
         self.dropout = nn.Dropout(dropout)
 
+        self.max_features = (
+            2 * ratio * output_features
+            if activation.__name__.endswith("GLU")
+            else ratio * output_features
+        )
+
+        if raw_input:
+            self.memory_type = AnchoredLinear
+
+        else:
+            self.memory_type = linear_module
+
         self.process = nn.Sequential(
             *[
                 nn.LayerNorm(input_features),
-                linear_module(
-                    input_features,
-                    (
-                        2 * ratio * output_features
-                        if activation.__name__.endswith("GLU")
-                        else ratio * output_features
-                    ),
-                ),
+                linear_module(input_features, self.max_features),
                 self.activation,
-                linear_module(ratio * output_features, output_features),
+                nn.LayerNorm(ratio * output_features) if raw_input else nn.Identity(),
+                self.memory_type(ratio * output_features, output_features),
                 self.dropout,
             ]
         )
@@ -296,7 +304,6 @@ class TransformerBlock(nn.Module):
 
         self.identity_probability = identity_probability
 
-        # Submodules for applying attention
         self.layer_norm = nn.LayerNorm(d_model)
 
         if position_embedding_type == "relative":
@@ -324,7 +331,7 @@ class TransformerBlock(nn.Module):
         )
 
         # Submodules for the feedforward process
-        self.ff = FeedforwardLayer(
+        self.ff = FeedforwardBlock(
             d_model,
             mlp_ratio,
             d_model,
