@@ -1,9 +1,10 @@
 import math
 from typing import Optional
 
-from .transformer import TransformerEncoder, FeedforwardLayer
+from .transformer import TransformerEncoder, FeedforwardBlock
 from .cnn import SpaceToDepth, calculate_output_spatial_size, spatial_tuple
 from .activation import ReLU, SquaredReLU, GELU, SwiGLU
+from .linear import SpectralNormLinear
 from einops import einsum
 from einops.layers.torch import Rearrange
 import torch.nn as nn
@@ -17,19 +18,23 @@ class PadTensor(nn.Module):
         self.kwargs = kwargs
 
     def forward(self, x):
-        return F.pad(x, *self.args, **self.kwargs)
+        if sum(self.args[0]) == 0:
+            return x
+        else:
+            return F.pad(x, *self.args, **self.kwargs)
+
+
+class GetCLSToken(nn.Module):
+    def __init__(self):
+        super().__init__()
+
+    def forward(self, x):
+        return x[:, 0, :]
 
 
 class SequencePool(nn.Module):
-    """
-    As described in [Hasani et al. (2021) *''Escaping the Big Data Paradigm with
-        Compact Transformers''*](https://arxiv.org/abs/2104.05704). It can be viewed
-        as a generalisation of average pooling.
-    """
-
-    def __init__(self, d_model, linear_module, out_dim, batch_norm=True):
+    def __init__(self, d_model, linear_module=nn.Linear):
         super().__init__()
-        self.d_model = d_model
         self.attention = nn.Sequential(
             *[
                 linear_module(d_model, 1),
@@ -37,20 +42,49 @@ class SequencePool(nn.Module):
                 nn.Softmax(dim=-1),
             ]
         )
-        self.projection = nn.Linear(d_model, out_dim)
-        self.batch_norm = batch_norm
-        if batch_norm:
-            self.norm = nn.BatchNorm1d(out_dim, affine=False)
-        else:
-            self.norm = None
 
     def forward(self, x):
         weights = self.attention(x)
-        weighted_embedding = einsum(
-            weights, x, "batch seq, batch seq d_model -> batch d_model"
+        return einsum(weights, x, "batch seq, batch seq d_model -> batch d_model")
+
+
+class ClassificationHead(nn.Module):
+    """
+    A general classification head for a ViT
+    """
+
+    def __init__(self, d_model, linear_module, n_classes, batch_norm=True):
+        super().__init__()
+        self.d_model = d_model
+        self.summarize = GetCLSToken()
+        self.projection = nn.Linear(d_model, n_classes)
+        if batch_norm:
+            self.batch_norm = nn.BatchNorm1d(n_classes, affine=False)
+        else:
+            self.batch_norm = nn.Identity()
+
+        self.classification_process = nn.Sequential(
+            *[
+                self.summarize,
+                self.projection,
+                self.batch_norm,
+            ]
         )
-        projection = self.projection(weighted_embedding)
-        return self.norm(projection) if self.batch_norm else projection
+
+    def forward(self, x):
+        return self.classification_process(x)
+
+
+class SequencePoolClassificationHead(ClassificationHead):
+    """
+    As described in [Hasani et al. (2021) *''Escaping the Big Data Paradigm with
+        Compact Transformers''*](https://arxiv.org/abs/2104.05704). It can be viewed
+        as a generalisation of average pooling.
+    """
+
+    def __init__(self, d_model, linear_module, out_dim, batch_norm=True):
+        super().__init__(d_model, linear_module, out_dim, batch_norm=batch_norm)
+        self.summarize = SequencePool(d_model, linear_module)
 
 
 class ViTEncoder(nn.Module):
@@ -66,8 +100,9 @@ class ViTEncoder(nn.Module):
     def __init__(
         self,
         input_size=(32, 32),
+        in_channels=3,
+        initial_batch_norm=True,
         cnn=True,
-        cnn_in_channels=3,
         cnn_out_channels=16,
         cnn_kernel_size=3,
         cnn_kernel_stride=1,
@@ -81,20 +116,20 @@ class ViTEncoder(nn.Module):
         pooling_kernel_size=3,
         pooling_kernel_stride=2,
         pooling_padding=1,
-        intermediate_feedforward_layer=True,
+        transformer_feedforward_first=True,
         transformer_position_embedding="relative",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
         transformer_heads=4,
         transformer_mlp_ratio=2,
         transformer_bos_tokens=0,
+        transformer_return_bos_tokens=False,
         transformer_activation: nn.Module = SquaredReLU,
         transformer_activation_kwargs: Optional[dict] = None,
         transformer_mlp_dropout=0.0,
         transformer_msa_dropout=0.1,
         transformer_stochastic_depth=0.1,
         linear_module=nn.Linear,
-        initial_batch_norm=True,
     ):
         super().__init__()
 
@@ -137,6 +172,9 @@ class ViTEncoder(nn.Module):
             )
 
         if cnn:
+            # This block rhymes:
+            if cnn_activation.__name__.endswith("GLU"):
+                cnn_out_channels *= 2
             cnn_output_size = calculate_output_spatial_size(
                 input_size,
                 kernel_size=cnn_kernel_size,
@@ -145,7 +183,7 @@ class ViTEncoder(nn.Module):
                 dilation=cnn_kernel_dilation,
             )
             self.cnn = convxd(
-                cnn_in_channels,
+                in_channels,
                 cnn_out_channels,
                 cnn_kernel_size,
                 stride=cnn_kernel_stride,
@@ -167,15 +205,12 @@ class ViTEncoder(nn.Module):
                     batchnormxd(cnn_activation_out_channels),
                 ]
             )
-            # This block rhymes:
-            if cnn and cnn_activation.__name__.endswith("GLU"):
-                cnn_out_channels *= 2
         else:
             self.cnn = nn.Identity()
             self.activate_and_dropout = nn.Identity()
             cnn_output_size = input_size
-            cnn_out_channels = cnn_in_channels
-            cnn_activation_out_channels = cnn_in_channels
+            cnn_out_channels = in_channels
+            cnn_activation_out_channels = in_channels
 
         pooling_kernel_voxels = math.prod(
             spatial_tuple(pooling_kernel_size, self.spatial_dimensions)
@@ -230,6 +265,10 @@ class ViTEncoder(nn.Module):
                 "Pooling type must be max, average, concat or None"
             )
 
+        self.pooling_channels_padding = PadTensor(
+            (0, max(0, transformer_embedding_size - pooling_out_channels))
+        )
+
         self.sequence_length = math.prod(pooling_output_size)  # One token per voxel
 
         if transformer_layers > 0:
@@ -249,39 +288,28 @@ class ViTEncoder(nn.Module):
                 causal=False,
                 linear_module=linear_module,
                 bos_tokens=transformer_bos_tokens,
+                return_bos_tokens=transformer_return_bos_tokens,
             )
         else:
             self.transformer = nn.Identity()
 
-        if intermediate_feedforward_layer:
-            self.pooling_channels_padding = nn.Identity()
-            self.intermediate_feedforward_layer = FeedforwardLayer(
-                pooling_out_channels,
+        if transformer_feedforward_first:
+            self.initial_ff = FeedforwardBlock(
+                max(transformer_embedding_size, pooling_out_channels),
                 transformer_mlp_ratio,
                 transformer_embedding_size,
                 activation=transformer_activation,
                 activation_kwargs=transformer_activation_kwargs,
                 dropout=transformer_mlp_dropout,
                 linear_module=linear_module,
-            )
-        elif pooling_out_channels < transformer_embedding_size:
-            self.intermediate_feedforward_layer = nn.Identity()
-            self.pooling_channels_padding = PadTensor(
-                (0, transformer_embedding_size - pooling_out_channels)
+                raw_input=not cnn,
             )
         else:
-            raise NotImplementedError(
-                "In a situation where the choice/parameters of the pooling and the"
-                + " `cnn_out_channels` (or the number of `input_channels` if"
-                + " `cnn`=False) means that the pooling will result"
-                + " in more channels per pixel/voxel than the size of the"
-                + " intended transformer embedding,"
-                + " `intermediate_feedforward_layer` must be set to True"
-            )
+            self.initial_ff = nn.Identity()
 
         self.encoder = nn.Sequential(
             *[
-                batchnormxd(cnn_in_channels) if initial_batch_norm else nn.Identity(),
+                batchnormxd(in_channels) if initial_batch_norm else nn.Identity(),
                 self.cnn,
                 self.activate_and_dropout,
                 self.pool,
@@ -289,7 +317,7 @@ class ViTEncoder(nn.Module):
                     f"N C {spatial_dim_names} -> N ({spatial_dim_names}) C"
                 ),
                 self.pooling_channels_padding,
-                self.intermediate_feedforward_layer,
+                self.initial_ff,
                 self.transformer,
             ]
         )
@@ -298,7 +326,7 @@ class ViTEncoder(nn.Module):
         return self.encoder(x)
 
 
-class CCT(nn.Module):
+class ViT(nn.Module):
     """
     Denoising convolutional transformer
     Based on the Compact Convolutional Transformer (CCT) of [Hasani et al. (2021)
@@ -310,8 +338,10 @@ class CCT(nn.Module):
     def __init__(
         self,
         input_size=(32, 32),
+        image_classes=100,
+        in_channels=3,
+        initial_batch_norm=True,
         cnn=True,
-        cnn_in_channels=3,
         cnn_out_channels=16,
         cnn_kernel_size=3,
         cnn_kernel_stride=1,
@@ -325,22 +355,22 @@ class CCT(nn.Module):
         pooling_kernel_size=3,
         pooling_kernel_stride=2,
         pooling_padding=1,
-        intermediate_feedforward_layer=True,
+        transformer_feedforward_first=True,
         transformer_position_embedding="relative",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
         transformer_heads=4,
         transformer_mlp_ratio=2,
         transformer_bos_tokens=0,
+        transformer_return_bos_tokens=False,
         transformer_activation: nn.Module = SquaredReLU,
         transformer_activation_kwargs: Optional[dict] = None,
         transformer_mlp_dropout=0.0,
         transformer_msa_dropout=0.1,
         transformer_stochastic_depth=0.1,
         batch_norm_outputs=True,
-        initial_batch_norm=True,
-        linear_module=nn.Linear,
-        image_classes=100,
+        linear_module=SpectralNormLinear,
+        head=SequencePoolClassificationHead,
     ):
 
         super().__init__()
@@ -363,8 +393,9 @@ class CCT(nn.Module):
 
         self.encoder = ViTEncoder(
             input_size=input_size,
+            initial_batch_norm=initial_batch_norm,
+            in_channels=in_channels,
             cnn=cnn,
-            cnn_in_channels=cnn_in_channels,
             cnn_out_channels=cnn_out_channels,
             cnn_kernel_size=cnn_kernel_size,
             cnn_kernel_stride=cnn_kernel_stride,
@@ -378,22 +409,23 @@ class CCT(nn.Module):
             pooling_kernel_size=pooling_kernel_size,
             pooling_kernel_stride=pooling_kernel_stride,
             pooling_padding=pooling_padding,
-            intermediate_feedforward_layer=intermediate_feedforward_layer,
+            transformer_feedforward_first=transformer_feedforward_first,
             transformer_position_embedding=transformer_position_embedding,
             transformer_embedding_size=transformer_embedding_size,
             transformer_layers=transformer_layers,
             transformer_heads=transformer_heads,
             transformer_mlp_ratio=transformer_mlp_ratio,
             transformer_bos_tokens=transformer_bos_tokens,
+            transformer_return_bos_tokens=transformer_return_bos_tokens,
             transformer_activation=transformer_activation,
             transformer_activation_kwargs=transformer_activation_kwargs,
             transformer_mlp_dropout=transformer_mlp_dropout,
             transformer_msa_dropout=transformer_msa_dropout,
             transformer_stochastic_depth=transformer_stochastic_depth,
             linear_module=linear_module,
-            initial_batch_norm=initial_batch_norm,
         )
-        self.pool = SequencePool(
+
+        self.pool = head(
             transformer_embedding_size,
             linear_module,
             image_classes,
